'''Context classes are the adaptors between data interfaces and templates.'''
from collections import UserDict, UserList
import logging
import os.path

from pygments import highlight
from pygments.lexers import YamlLexer
from pygments.formatters import HtmlFormatter


class AccordionDataContext(UserDict):
    def __init__(self, name, iterable):
        initial = {
            'cssid': name.lower(),
            'name': name,
            'iterable': iterable,
        }
        super().__init__(initial)


class HighlightedYamlContext(UserDict):
    def __init__(self, initial):
        content = highlight(initial.as_yaml(), YamlLexer(), HtmlFormatter())
        initial['yaml_highlight_content'] = content
        super().__init__(initial)


class MachinesContext(UserList):
    @property
    def notrunning(self):
        ret = list(filter(lambda m: m.get('status', {}).get('phase') != 'Running', self.data))
        return ret


class MachineSetContext(HighlightedYamlContext):
    @property
    def autoscaler_min(self):
        '''return autoscaler min or none'''
        return self.data.get('metadata', {}).get('annotations', {}).get('machine.openshift.io/cluster-api-autoscaler-node-group-min-size')

    @property
    def autoscaler_max(self):
        '''return autoscaler max or none'''
        return self.data.get('metadata', {}).get('annotations', {}).get('machine.openshift.io/cluster-api-autoscaler-node-group-max-size')


class NodesContext(UserList):
    @property
    def notready(self):
        notready = []
        for node in self.data:
            for condition in node.get('status', {}).get('conditions', []):
                if condition.get('type') == 'Ready' and condition.get('status') == 'False':
                    notready.append(node)
        return notready


class PodContext(HighlightedYamlContext):
    def __init__(self, pod):
        super().__init__(pod)
        self.data['containerlogs'] = [{'name': k, 'logs': v} for k, v in pod.containerlogs.items()]


class ResourceContext(HighlightedYamlContext):
    pass


class NavListContext(UserDict):
    def __init__(self, cssid, anchor_name, content):
        initial = {
            'id': cssid,
            'anchor_name': anchor_name,
            'content': content,
        }
        super().__init__(initial)


class IndexContext(UserDict):
    '''Context for the index.html template'''
    def __init__(self, mustgather):
        mapipods = [PodContext(pod) for pod in mustgather.pods('openshift-machine-api')]
        machineautoscalers = [ResourceContext(machineautoscaler) for machineautoscaler in mustgather.machineautoscalers]
        clusterautoscalers = [ResourceContext(clusterautoscaler) for clusterautoscaler in mustgather.clusterautoscalers]
        machines = MachinesContext([ResourceContext(machine) for machine in mustgather.machines])
        nodes = NodesContext([ResourceContext(node) for node in mustgather.nodes])
        initial = {
            'accordiondata': [
                AccordionDataContext('ClusterAutoscalers', clusterautoscalers),
                AccordionDataContext('MachineAutoscalers', machineautoscalers),
                AccordionDataContext('Machines', machines),
                AccordionDataContext('Nodes', nodes),
            ],
            'basename': self.basename(mustgather.path),
            'clusterautoscalers': clusterautoscalers,
            'highlight_css': HtmlFormatter().get_style_defs('.highlight'),
            'machineautoscalers': machineautoscalers,
            'machines': machines,
            'machinesets': [MachineSetContext(machineset) for machineset in mustgather.machinesets],
            'machinesets_participating': [ msc for msc in [MachineSetContext(machineset) for machineset in mustgather.machinesets] if msc.autoscaler_min],
            'mapipods': mapipods,
            'nodes': nodes,
        }
        super().__init__(initial)

    @staticmethod
    def basename(path):
        if path.endswith('/'):
            path = path[:-1]
        return os.path.basename(path)

    @staticmethod
    def cluster_autoscaler_deployment(mustgather):
        deployment = mustgather.clusterautoscaler.deployment
        if deployment is None:
            return 'Deployment not found, check <must-gather path>/namespaces/openshift-machine-api/apps/deployments.yaml'
        content = highlight(deployment.as_yaml(), YamlLexer(), HtmlFormatter())
        anchor = deployment.name()
        return NavListContext(cssid='cluster-autoscaler-deployment', anchor_name=anchor, content=content)

    @staticmethod
    def cluster_autoscaler_pods(mustgather):
        ret = []
        for pod in  mustgather.clusterautoscaler.pods:
            name = pod.name()
            content = highlight(pod.as_yaml(), YamlLexer(), HtmlFormatter())
            ret.append(NavListContext(cssid=name, anchor_name=name, content=content))
        return ret




from argparse import ArgumentParser
import logging
import os.path
from tempfile import mkdtemp
from threading import Thread
from time import sleep
import webbrowser

from bottle import route, run
from jinja2 import Environment, PackageLoader, select_autoescape

import okd_camgi
from okd_camgi.contexts import IndexContext
from okd_camgi.interfaces import MustGather


def load_index_from_path(path):
    env = Environment(
        loader=PackageLoader('okd_camgi', 'templates'),
        autoescape=False
    )

    mustgather = MustGather(path)

    # render the index.html template
    index_template = env.get_template('index.html')
    index_context = IndexContext(mustgather)
    index_content = index_template.render(index_context.data)

    return index_content


def main():
    logging.basicConfig(level=logging.DEBUG)
    parser = ArgumentParser(prog='okd-camgi', description='investigate a must-gather for clues of autoscaler activity')
    parser.add_argument('path', help='path to the root of must-gather tree')
    parser.add_argument('--webbrowser', action='store_true', help='open a webbrowser to investigation')
    parser.add_argument('--server', action='store_true', help='run in server mode')
    parser.add_argument('--host', help='server host address', default='127.0.0.1')
    parser.add_argument('--port', help='server host port', default='8080')
    parser.add_argument('--output', help='output filename')
    parser.add_argument('--version', action='version', version=f'%(prog)s {okd_camgi.version}')
    args = parser.parse_args()

    content = load_index_from_path(args.path)
    if args.output or not args.server:
        indexpath = args.output if args.output else os.path.join(mkdtemp(), 'index.html')
        with open(indexpath, 'w') as indexfile:
            indexfile.write(content)

    host = args.host
    port = int(args.port)
    url = f'file://{indexpath}' if not args.server else f'http://{host}:{port}/'
    print(f'{url}')

    bth = None
    if args.webbrowser:
        # delay opening the browser in case we are running in server mode
        def delay_browser_open():
            sleep(1)
            webbrowser.open(url)

        bth = Thread(target=delay_browser_open)
        bth.start()

    if args.server:
        @route('/')
        def handler():
            content = load_index_from_path(args.path)
            return content

        run(host=host, port=port, debug=True)

    if bth is not None:
        bth.join()


if __name__ == '__main__':
    main()



'''Interfaces into the must gather artifacts and data.'''
from collections import UserDict
from copy import deepcopy
import json
import logging
import os.path

import yaml


class Resource(UserDict):
    def name(self):
        return self.data.get('metadata', {}).get('name')

    def as_yaml(self):
        data = deepcopy(self.data)
        if data.get('metadata', {}).get('managedFields'):
            del data['metadata']['managedFields']
        return yaml.dump(data)


class Pod(Resource):
    def __init__(self, resource, containerlogs):
        super().__init__(resource)
        self.containerlogs = containerlogs


class MustGather:
    def __init__(self, path):
        self.path = path
        self._clusterautoscalers = None
        self._machineautoscalers = None
        self._machines = None
        self._machinesets = None
        self._nodes = None
        self._pods = {}

    @property
    def clusterautoscalers(self):
        if self._clusterautoscalers is None:
            self._clusterautoscalers = self.resources('clusterautoscalers', 'autoscaling.openshift.io')
        return self._clusterautoscalers

    @property
    def machineautoscalers(self):
        if self._machineautoscalers is None:
            self._machineautoscalers = self.resources('machineautoscalers', 'autoscaling.openshift.io', 'openshift-machine-api')
        return self._machineautoscalers

    @property
    def machines(self):
        if self._machines is None:
            machines = self.resources('machines', 'machine.openshift.io', 'openshift-machine-api')
            self._machines = sorted(machines, key=lambda m: m.name())
        return self._machines

    @property
    def machinesets(self):
        if self._machinesets is None:
            machinesets = self.resources('machinesets', 'machine.openshift.io', 'openshift-machine-api')
            self._machinesets = sorted(machinesets, key=lambda m: m.name())
        return self._machinesets

    @property
    def nodes(self):
        if self._nodes is None:
            nodes = self.resources('nodes', 'core')
            self._nodes = sorted(nodes, key=lambda n: n.name())
        return self._nodes

    def pods(self, namespace):
        if self._pods.get(namespace) is None:
            pods = []
            pods_path = self.build_manifest_path(self.path, None, 'pods', None, namespace)
            # list all pods in the namespace
            for podname in os.listdir(pods_path):
                containerlogs = {}
                # list all files in the pod dir (containers and manifest)
                for filename in os.listdir(os.path.join(pods_path, podname)):
                    resource = None
                    if filename == f'{podname}.yaml':
                        logging.debug(f'loading {filename}')
                        with open(os.path.join(pods_path, podname, filename)) as man_file:
                            resource = yaml.load(man_file.read(), Loader=yaml.FullLoader)
                    # if this is a container, see if there are logs
                    elif os.path.exists(os.path.join(pods_path, podname, filename, filename, 'logs')):
                        for logfile in os.listdir(os.path.join(pods_path, podname, filename, filename, 'logs')):
                            logging.debug(f'found container logs for {filename}, opening {logfile}')
                            with open(os.path.join(pods_path, podname, filename, filename, 'logs', logfile)) as log:
                                containerlogs[logfile] = log.read()
                    else:
                        continue
                    if resource is not None:
                        pods.append(Pod(resource, containerlogs))
            self._pods[namespace] = pods
        return self._pods[namespace]

    @staticmethod
    def build_manifest_path(path, name, kind, group, namespace):
        pathlist = [path]
        if namespace is None:
            pathlist.append('cluster-scoped-resources')
        else:
            pathlist.append('namespaces')
            pathlist.append(namespace)
        if group is not None:
            pathlist.append(group)
        pathlist.append(kind)
        if name is not None:
            pathlist.append(f'{name}.yaml')
        man_path = os.path.join(*pathlist)
        return man_path

    def resource_or_none(self, name, kind, group=None, namespace=None):
        '''get a resource or none if not found'''
        man_path = self.build_manifest_path(self.path, name, kind, group, namespace)

        if not os.path.exists(man_path):
            return None
        logging.debug(f'loading {group}/{kind} yaml from {man_path}')
        with open(man_path) as man_file:
            resource = yaml.load(man_file.read(), Loader=yaml.FullLoader)
            return Resource(resource)

    def resources(self, kind, group=None, namespace=None):
        yaml_path = self.build_manifest_path(self.path, None, kind, group, namespace)
        resourcelist = []
        if not os.path.exists(yaml_path):
            return resourcelist
        for f in os.listdir(yaml_path):
            if not f.endswith('.yaml'):
                continue
            resource = self.resource_or_none(f[:-5], kind, group, namespace)
            if resource is None:
                logging.error(f'Found yaml {f} did not produce a resource.')
            else:
                resourcelist.append(resource)
        return resourcelist



from setuptools import setup

import okd_camgi


with open('README.md', 'r', encoding='utf-8') as readmefile:
    readme = readmefile.read()


dependencies = [
    'jinja2',
    'bottle',
    'pyyaml',
    'pygments',
]


setup(
    name='okd-camgi',
    version=okd_camgi.version,
    license='GPLv3',
    author='michael mccune',
    author_email='msm@opbstudios.com',
    description='OKD Cluster Autoscaler Must Gather Investigator',
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/elmiko/okd-camgi",
    packages=['okd_camgi'],
    install_requires=dependencies,
    package_data={
        'okd_camgi': ["templates/*.html"],
    },
    entry_points={
        'console_scripts': [
            'okd-camgi = okd_camgi.main:main',
        ],
    }
)
